# -*- coding: utf-8 -*-
# @Author: Jason Y. Wu
# @Date:   2023-08-04 13:59:52
# @Last Modified by:   Jason Y. Wu
# @Last Modified time: 2023-08-06 20:52:08
from typing import List, Dict, Union, Tuple, Callable

from gailbot.services import ServiceController, SettingDict
from gailbot.workspace import WorkspaceManager
from .plugins.suite import PluginSuite
from gailbot.core.utils.logger import makelogger

logger = makelogger("gb_api")


class GailBot:
    """
    Class for API wrapper
    """

    def __init__(self, ws_root: str):
        """initialize a gailbot object that provides a suite of functions
            to interact with gailbot

        Args:
            ws_root (str): the path to workspace root
        """
        self.ws_manager: WorkspaceManager = WorkspaceManager(ws_root)
        self.init_workspace()
        logger.info("workspace manager initialized")
        self.gb: ServiceController = ServiceController(
            self.ws_manager, load_exist_setting=True
        )
        logger.info("gailbot service controller initialized")

    def init_workspace(self):
        """
        Resets the workspace: clears the old workspace and initializes a new one.

        Returns:
            No return but instantiates a new workspace.
        """
        try:
            self.ws_manager.clear_gb_temp_dir()
            self.ws_manager.init_workspace()
            return True
        except Exception as e:
            logger.error(f"failed to reset workspace due to the error {e}", exc_info=e)
            return False

    def clear_workspace(self):
        """
        Clears current workspace

        Returns: None
        """
        try:
            self.ws_manager.clear_gb_temp_dir()
            return True
        except Exception as e:
            logger.error(f"failed to reset workspace due to the error {e}", exc_info=e)
            return False

    def reset_workspace(self) -> bool:
        """
        Reset the gailbot workspace

        Returns: True if workspace successfully reset; false otherwise
        """
        return self.ws_manager.reset_workspace()

    ###########################################################################
    # Sources                                                                 #
    ###########################################################################

    def add_source(self, source_path: str, output_dir: str) -> bool:
        """
        Adds a given source

        Args:
            source_path : str: Source path of the given source
            output_dir : str: Path to the output directory of the given source

        Returns:
            Union[str, bool]: return the name if successfully added, false if not
        """
        return self.gb.add_source(source_path, output_dir)

    def add_sources(self, src_output_pairs: List[Tuple[str, str]]) -> bool:
        """
        Adds a given list of sources

        Args:
            src_output_pairs: List [Tuple [str, str]]: List of Tuples of strings
                to strings, each representing the source path and output path of
                a source to add

        Returns:
            Bool: True if each given source was successfully added, false if not
        """
        return self.gb.add_sources(src_output_pairs)

    def is_source(self, name: str) -> bool:
        """
        Determines if a given name corresponds to an existing source

        Args:
            name: str: Name of the source to look for

        Returns:
            Bool: True if the given name corresponds to an existing source,
                  false if not
        """
        return self.gb.is_source(name)

    def get_source_outdir(self, name: str) -> str:
        """
        Accesses source output directory with a given name

        Args:
            source_name: str: source name to access

        Returns:
            a string stores the output path of the source
        """
        return self.gb.get_source_out_dir(name)

    def remove_source(self, source_name: str) -> bool:
        """
        Removes the given source

        Args:
            source_name : str: Name of the existing source to remove

        Returns:
            Bool: True if source was successfully removed, false if not
        """
        return self.gb.remove_source(source_name)

    def remove_sources(self, source_names: List[str]) -> bool:
        """
        Removes the given list of sources
        Args:
            source_name : str: Name of the existing sources to remove
        Returns:
            Bool: True if all sources were successfully removed, false if not
        """
        return self.gb.remove_sources(source_names)

    def get_source_setting_dict(self, source_name) -> Union[bool, SettingDict]:
        """
        Given a source, returns its setting content as a dictionary

        Args:
            source_name (str): the name of the source

        Returns:
            Union[bool, Dict[str, Union[str, Dict]]]:  a dictionary that stores
            the source setting content
        """
        return self.gb.get_source_setting_dict(source_name)

    def clear_source_memory(self) -> bool:
        """
        Clears source memory

        Returns:
            bool: True if successfully cleared, False if not
        """
        return self.gb.clear_source_memory()

    def get_all_source_names(self) -> List[str]:
        """
        Returns list of all source names

        Returns: List[str] : list of source names
        """
        return self.gb.get_all_source_names()

    def get_src_setting_name(self, source_name: str) -> Union[bool, str]:
        """given a source name, return the setting name applied to the source

        Args:
            source_name (str): the name that identify the source

        Returns:
            Union[bool, str]: if the source is found, return the setting name
                              applied to the source, else return false
        """
        return self.gb.get_src_setting_name(source_name)

    def get_source_setting_dict(self, source_name) -> Union[bool, SettingDict]:
        """
        Given a source, returns its setting content as a dictionary

        Args:
            source_name (str): the name of the source

        Returns:
            Union[bool, Dict[str, Union[str, Dict]]]:  a dictionary that stores
            the source setting content
        """
        return self.gb.get_source_setting_dict(source_name)

    ###########################################################################
    # Transcribe                                                              #
    ###########################################################################

    # def keep
    def transcribe(self, sources: List[str] = None) -> Tuple[List[str], List[str]]:
        """given a list of the source name, and transcribe the sources

        Args:
            sources (List[str], optional): a list of source name, which
            can be either a list of source paths or the file name of the
            source file without the file extension
            if sources is None, the default is to transcribe all sources
            that have been configured

        Returns:
            Tuple[bool, List[str]]:
                returns a tuple of two lists of string
                the first lists consist of files that are not valid input
                the second lists consist of files that fails to be processed

        """
        return self.gb.transcribe(sources)

    ###########################################################################
    # Setting (Profile)                                                       #
    ###########################################################################

    def create_new_setting(
        self, name: str, setting: Dict[str, str], overwrite: bool = True
    ) -> bool:
        """
        Creates a new setting profile

        Args:
            name: str: Name to assign to the newly created setting profile
            setting : Dict[str, str]: Dictionary representation of the setting
            overwrite : bool

        Returns:
            True if setting was successfully created, false if not
        """
        return self.gb.create_new_setting(name, setting, overwrite)

    def save_setting(self, setting_name: str) -> bool:
        """
        Saves the given setting

        Args:
            setting_name: str: Name of the setting to save

        Returns:
            Bool: True if setting was successfully saved, false if not
        """
        return self.gb.save_setting(setting_name)

    def get_setting_dict(self, setting_name: str) -> Union[bool, SettingDict]:
        """
        Given a setting name, returns the setting content in a dictionary

        Args:
            setting_name (str): name that identifies a setting

        Returns:
            Union[bool, SettingDict]: if the setting is found, returns its setting
            content stored in a dictionary, else returns false
        """
        return self.gb.get_setting_dict(setting_name)

    def get_all_settings_data(self) -> Dict[str, SettingDict]:
        """
        given a setting name, return the setting content in a dictionary format

        Returns:
            Union[bool, Dict[str, Union[str, Dict]]]: the setting content
        """
        return self.gb.get_all_settings_data()

    def get_all_profile_names(self) -> List[str]:
        """get the names fo available settings

        Returns:
            List[str]: a list of available setting names
        """
        return self.gb.get_all_settings_names()

    def rename_setting(self, old_name: str, new_name: str) -> bool:
        """
        Renames a given setting to a given new name

        Args:
            old_name: str: original name of the setting to rename
            new_name: str: name to rename the setting to

        Returns:
            Bool: True if setting was successfully renamed, false if not

        """
        return self.gb.rename_setting(old_name, new_name)

    def update_setting(self, setting_name: str, new_setting: Dict[str, str]) -> bool:
        """
        Updates a given setting to a newly given structure

        Args:
            setting_name: str: name of the setting to update
            new_setting: Dict[str, str]: dictionary representation of
                the new structure of the setting

        Returns:
            Bool: true if setting was successfully updated, false if not
        """
        return self.gb.update_setting(setting_name, new_setting)

    # keep
    def get_plugin_setting(self, setting_name: str) -> bool | list[str]:
        """
        Accesses the plugin setting of a given setting

        Args:
            setting_name: str: name of the setting to get the plugin setting of

        Returns:
            Dict[str, str]: dictionary representation of the plugin setting
        """
        return self.gb.get_plugin_setting(setting_name)

    def remove_setting(self, setting_name: str) -> bool:
        """
        Removes the given setting

        Args:
            setting_name: str: name of the setting to remove

        Returns:
            Bool: True if setting was successfully removed, false if not
        """
        return self.gb.remove_setting(setting_name)

    def remove_multiple_settings(self, setting_names: List[str]) -> bool:
        """
        Removes the given list of settings

        Args:
            setting_name: List[str]: names of the setting to remove

        Returns:
            Bool: True if all settings were successfully removed, false if not
        """
        return self.gb.remove_multiple_settings(setting_names)

    def is_setting(self, name: str) -> bool:
        """
        Determines if a given setting name corresponds to an existing setting

        Args:
            name: str: name of the setting to search fort

        Returns:
            Bool: True if given setting is an existing setting, false if not
        """
        return self.gb.is_setting(name)

    def apply_setting_to_source(
        self, source: str, setting: str, overwrite: bool = True
    ) -> bool:
        """
        Applies a given setting to a given source

        Args:
            source: str: name of the source to which to apply the given setting profile
            setting: str: name of the setting to apply to the given source
            overwrite: bool: determines if it should overwrite from an existing setting
                Defaults to true

        Returns:
            Bool: true if setting was successfully applied, false if not
        """
        return self.gb.apply_setting_to_source(source, setting, overwrite)

    def apply_setting_to_sources(
        self, sources: List[str], setting: str, overwrite: bool = True
    ) -> bool:
        """
        Applies a given setting to a given list of sources

        Args:
            sources: List[str]: list of names of the sources to which to apply the given setting profile
            setting: str: name of the setting to apply to the given sources
            overwrite: bool: determines if it should overwrite from an existing setting
                Defaults to true

        Returns:
            Bool: true if setting was successfully applied, false if not

        """
        return self.gb.apply_setting_to_sources(sources, setting, overwrite)

    def is_setting_in_use(self, setting_name: str) -> bool:
        """check if a setting is being used by any source

        Args:
            setting_name (str): the name of the setting

        Returns:
            bool: return true if the setting is being used, false otherwise
        """
        return self.gb.is_setting_in_use(setting_name)

    def get_default_profile_setting_name(self) -> str:
        """get the name of current default setting

        Returns:
            str: the name of current default setting
        """
        return self.gb.get_default_profile_setting_name()

    def get_default_engine_setting_name(self) -> str:
        """get the default engine setting name

        Returns:
            str: a string that represent the default engine setting
        """
        return self.gb.get_default_engine_setting_name()

    def set_default_setting(self, setting_name) -> bool:
        """set the default setting to setting name

        Args:
            setting_name (str): the name of the default setting

        Returns:
            bool: true if default setting is set correctly
        """
        return self.gb.set_default_setting(setting_name)

    ###########################################################################
    # Plugin Suite                                                            #
    ###########################################################################

    def register_plugin_suite(self, plugin_source: str) -> Union[List[str], str]:
        """
        Registers a gailbot plugin suite

        Args:
            plugin_source : str: Name of the plugin suite to register

        Returns:
            return  a list of plugin name if the plugin is registered,
            return the string that stores the error message if the plugin suite
            is not registered
        """
        return self.gb.register_plugin_suite(plugin_source)

    def get_plugin_suite(self, suite_name) -> PluginSuite:
        """
        Gets the plugin suite with a given name

        Args:
            suite_name: string name of the given plugin suite

        Returns:
            PluginSuite with the given name
        """
        return self.gb.get_plugin_suite(suite_name)

    def is_plugin_suite(self, suite_name: str) -> bool:
        """
        Determines if a given plugin suite is an existing plugin suite

        Args:
            suite_name: str: name of the plugin suite of which to determine existence

        Returns:
            Bool: true if given plugin suite exists, false if not
        """
        return self.gb.is_plugin_suite(suite_name)

    def delete_plugin_suite(self, suite_name: str) -> bool:
        """
        Removes the given plugin suite

        Args:
            suite_name: str: name of the plugin suite to delete

        Returns:
            Bool: true if plugin suite was successfully removed, false if not
        """
        return self.gb.delete_plugin_suite(suite_name)

    def delete_plugin_suites(self, suite_names: List[str]) -> bool:
        """
        Removes the given list of plugin suites

        Args:
            suite_name: List[str]: list of names of the plugin suites to delete

        Returns:
            Bool: true if all plugin suites were successfully removed, false if not
        """
        return self.gb.delete_plugin_suites(suite_names)

    def add_progress_display(self, source: str, displayer: Callable) -> bool:
        """
        Add a function displayer to track for the progress of source,

        Args:
            source (str): the name of the source
            displayer (Callable): displayer is a function that takes in a string as
                                  argument, and the string encodes the progress of
                                  the source

        Returns:
            bool: return true if the displayer is added, false otherwise
        """
        return self.gb.add_progress_display(source, displayer)

    def get_all_plugin_suites(self) -> List[str]:
        """get names of available plugin suites

        Returns:
            List[str]: a list of available plugin suites name
        """
        return self.gb.get_all_plugin_suites()

    def get_plugin_suite_metadata(self, suite_name: str) -> Dict[str, str]:
        """get the metadata of a plugin suite identified by suite name

        Args:
            suite_name (str): the name of the suite

        Returns:
            MetaData: a MetaData object that stores the suite's metadata,

        """
        return self.gb.get_plugin_suite_metadata(suite_name)

    def get_plugin_suite_dependency_graph(
        self, suite_name: str
    ) -> Dict[str, List[str]]:
        """get the dependency map of the plugin suite identified by suite_name

        Args:
            suite_name (str): the name of the suite

        Returns:
            Dict[str, List[str]]: the dependency graph of the suite
        """
        return self.gb.get_plugin_suite_dependency_graph(suite_name)

    def get_plugin_suite_documentation_path(self, suite_name: str) -> str:
        """get the path to the documentation map of the plugin suite identified by suite_name

        Args:
            suite_name (str): the name of the suite

        Returns:
            str: the path to the documentation file
        """
        return self.gb.get_plugin_suite_documentation_path(suite_name)

    def is_suite_in_use(self, suite_name: str) -> bool:
        """given a suite_name, check if this suite is used
           in any of the setting

        Args:
            suite_name (str): the name of the plugin suite

        Returns:
            bool: return true if the suite is used in any of the setting,
                  false otherwise
        """
        return self.gb.is_suite_in_use(suite_name)

    def is_official_suite(self, suite_name: str) -> bool:
        """given a suite_name, check if the suite identified by the suite_name
           is official

        Args:
            suite_name (str): the name of the suite

        Returns:
            bool: true if the suite is official false otherwise
        """
        return self.gb.is_official_suite(suite_name)

    def get_suite_source_path(self, suite_name: str) -> str:
        """
        given the name of the  suite , return the path to the source
        code of the suite
        """
        return self.gb.get_suite_path(suite_name)

    ###########################################################################
    # Engines                                                                 #
    ###########################################################################

    def get_engine_setting_names(self) -> List[str]:
        """get a list of available engine setting name

        Returns:
            List[str]: the list of engine setting name
        """
        return self.gb.get_engine_setting_names()

    def add_new_engine(self, name, setting, overwrite=False) -> bool:
        """add a new engine setting

        Args:
            name (str): the name of the engine setting
            setting (Dict[str, str]): the setting data stored in a dictionary
            overwrite (bool, optional): if True, overwrite the existing
                                        engine setting with the same name. Defaults to False.

        Returns:
            bool: return True if the engine setting is successfully created
        """
        return self.gb.add_new_engine(name, setting, overwrite)

    def remove_engine_setting(self, name) -> bool:
        """remove the engine setting identified by name

        Args:
            name (str): the name of the engine setting to be removed

        Returns:
            bool:  return True if the engine setting is successfully removed
        """
        return self.gb.remove_engine_setting(name)

    def update_engine_setting(self, name, setting_data: Dict[str, str]) -> bool:
        """update the engine setting identified by name

        Args:
            name (str): the name of the engine setting to be updated
            setting_data (Dict[str, str]): the content of the new setting

        Returns:
            bool:  return True if the engine setting is successfully updated
        """
        return self.gb.update_engine_setting(name, setting_data)

    def get_engine_setting_data(self, name: str) -> Union[bool, Dict[str, str]]:
        """get the engine setting data

        Args:
            name (str): the name of the engine setting

        Returns:
            Union[bool, Dict[str, str]]: if the engine setting name is available
            return the engine setting data as stored in a dictionary, else return False
        """
        return self.gb.get_engine_setting_data(name)

    def is_engine_setting_in_use(self, name: str) -> bool:
        """check if the engine setting identified by name is in use

        Args:
            name (str): the name of the engine setting

        Returns:
            bool: return true if the engine setting is in use, false otherwise
        """
        return self.gb.is_engine_setting_in_use(name)

    def is_engine_setting(self, name: str):
        """check if the given engine name is engine setting

        Args:
            name (str): the name of the engine setting
        """
        return self.gb.is_engine_setting(name)
